/*  Alien Pool
    Copyright (C) 2003  Michael J. Korman <mike@taequin.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/**
 * @file Alien.cc
 * Implementation of the Alien class.
 */

#include <cmath>
#include <list>

#include "Alien.hh"
#include "Level.hh"
#include "Bullet.hh"
#include "Ship.hh"
#include "common.hh"

extern Bullet** alien_bullets;
extern std::list<Level>::iterator current_level;
extern Ship* ship;
extern int level_num;

/**
 * Constructor.
 * @param base1 the sprite base of the normal alien.
 * @param base2 the sprite base of the dumb alien.
 * @param scr the screen surface.
 * @param bckgrnd the background surface.
 * @param xvel the initial x velocity
 * @param yvel the initial y velocity.
 * @param angle the initial angle.
 */
Alien::Alien(SpriteBase* base1, SpriteBase* base2, SDL_Surface* scr, SDL_Surface* bckgrnd, 
             float xvel, float yvel, int angle)
    : Sprite(base1, scr, bckgrnd, xvel, yvel, angle)
{
    _shoot_counter = rand() % (current_level->alien_shoot_time * 2);
    _angle = (rand() % 360) + 1;
    _base1 = base1;
    _base2 = base2;
    _state = ALIEN_NORMAL;
    _xvel = rand() % int(xvel) + 1;
    _yvel = rand() % int(yvel) + 1;
}

/**
 * Destructor. Updates the background.
 */
Alien::~Alien()
{
    update_background();
}

/**
 * If the alien is normal, check to see if we need to shoot, and then move one
 * step. If the alien is dumb, move according to our trajectory. If the dumb 
 * counter has reached 0, then kill the alien.
 * @return @b true if the alien is still alive, @b false otherwise.
 */
bool Alien::go()
{
    if (_state == ALIEN_NORMAL) {
        _shoot_counter--;

        if (_shoot_counter < 0) {
            shoot();
            _shoot_counter = rand() % current_level->alien_shoot_time;
        }

        int x = _x;
        int y = _y;
        
        int rand_x = _xvel;
        int rand_y = _yvel;
            
        if (ship->get_x() < _x) {
            x = _x - rand_x;
        } else if (ship->get_x() > _x) {
            x = _x + rand_x;
        }
            
        if (ship->get_y() < _y) {
            y = _y - rand_y;
        } else if (ship->get_y() > _y) {
            y = _y + rand_y;
        }

        move(x, y);
    
        return true;
    } else {
        _dumb_counter--;

        if (_dumb_counter == 0) {
            return false;
        } else {
            int new_x = _x + int(round_to_int(_xvel * cos((M_PI/180) * _trajectory)));
            int new_y = _y - int(round_to_int(_yvel * sin((M_PI/180) * _trajectory)));
            
            move(new_x, new_y);
            
            if (new_y < 0 || new_y > (SCREEN_HEIGHT - _sprite_base->get_height())) {
                _trajectory *= -1;
            }
            if (new_x > (SCREEN_WIDTH - _sprite_base->get_height()) || new_x < 0) {
                _trajectory = 180 - _trajectory;
            }
        }

        return true;
    }
}

/**
 * Shoots an alien bullet.
 */
void Alien::shoot()
{
    int i;
    Bullet* bullet = new Bullet(_bullet_base, _screen, _background, BULLET_CHANGE, BULLET_CHANGE, 0);
    
    for (i = 0; i < NUM_ALIEN_BULLETS; i++) {
	if (alien_bullets[i] == NULL) {
	    bullet->set_x(_x + (_sprite_base->get_width() / 2));
	    bullet->set_y(_y + (_sprite_base->get_height() / 2));
            bullet->set_trajectory(rand() % 360);
	    alien_bullets[i] = bullet;
	    break;
	}
    }
}

/**
 * Called when an alien is hit by a player bullet. Changes the sprite base, 
 * then makes the alien dumb.
 */
void Alien::hit()
{
    update_background();
    _sprite_base = _base2;
    _dumb_counter = current_level->alien_dumb_time;
    _state = ALIEN_DUMB;
    _xvel = current_level->dumbalien_change;
    _yvel = current_level->dumbalien_change;
}

/**
 * Gets the normal alien base.
 * @return the normal alien base.
 */
const SpriteBase* Alien::get_base1()
{
    return _base1;
}

/**
 * Gets the dumb alien base.
 * @return the dumb alien base.
 */
const SpriteBase* Alien::get_base2()
{
    return _base2;
}

/**
 * Gets the alien state.
 * @return the alien state.
 */
const ALIEN_STATE Alien::get_state()
{
    return _state;
}

/**
 * Gets the trajectory.
 * @return the trajectory;
 */
const int Alien::get_trajectory()
{
    return _trajectory;
}

/**
 * Gets the bullet base.
 * @return the bullet base.
 */
const SpriteBase* Alien::get_bullet_base()
{
    return _bullet_base;
}

/**
 * Sets the normal alien base.
 * @param base1 the new normal alien base.
 */
void Alien::set_base1(SpriteBase* base1)
{
    _base1 = base1;
}

/**
 * Sets the dumb alien base.
 * @param base2 the new dumb alien base.
 */
void Alien::set_base2(SpriteBase* base2)
{
    _base2 = base2;
}

/**
 * Sets the alien state.
 * @param the new alien state.
 */
void Alien::set_state(ALIEN_STATE state)
{
    _state = state;
}

/**
 * Sets the trajectory.
 * @param trajectory the new trajectory.
 */
void Alien::set_trajectory(int trajectory)
{
    _trajectory = trajectory;
}

/**
 * Sets the bullet base.
 * @param bullet_base the new bullet base.
 */
void Alien::set_bullet_base(SpriteBase* bullet_base)
{
    _bullet_base = bullet_base;
}
