/*  Alien Pool
    Copyright (C) 2003  Michael J. Korman <mike@taequin.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/**
 * @file Ship.cc
 * Implementation of the Ship class.
 */

#include <iostream>

#include "Ship.hh"

/**
 * Constructor. Sets the initial health and makes the player immune.
 * @param sb the sprite base.
 * @param scr the screen surface.
 * @param bckgrnd the background surface.
 * @param xvel the initial x velocity.
 * @param yvel the initial y velocity.
 * @param angle the initial angle.
 */
Ship::Ship(SpriteBase* base1, SpriteBase* base2, SDL_Surface* scr, SDL_Surface* bckgrnd, 
           float xvel, float yvel, int angle)
    : Sprite(base1, scr, bckgrnd, xvel, yvel, angle)
{
    _health = PLAYER_HEALTH;
    _state = 0;
    _bullets = 0;
    _max_bullets = 1;
    _score = 0;
    _base1 = base1;
    _base2 = base2;
}

/**
 * Destructor.
 */
Ship::~Ship()
{}

/**
 * Called to inflict damage upon the ship. If the ship is normal (not immune),
 * decrements the health and makes the ship immune. If health reaches 0,
 * signal a game over.
 * @param damage the damage to be inflicted.
 */
void Ship::hit(int damage)
{
    if (_state & PLAYER_IMMUNE) {
        return;
    } else {
        _health -= damage;
        make_immune();

        if (_health <= 0) {
            game_over();
        }
    }
}

/**
 * Updates the state of the bonuses.
 */
void Ship::go()
{
    _immune_counter--;
    _shoot_counter--;
    _hyper_counter--;

    if (_immune_counter < 0) {
        make_unimmune();
    }
    if (_shoot_counter < 0) {
        make_unshoot();
    }
    if (_hyper_counter < 0) {
        make_unhyper();
    }
}

/**
 * Makes the ship immune.
 */
void Ship::make_immune()
{
    _immune_counter = PLAYER_IMMUNE_TIME;
    _state |= PLAYER_IMMUNE;
}

/**
 * Makes the ship no longer immune.
 */
void Ship::make_unimmune()
{
    _state = _state & ~(PLAYER_IMMUNE);
}

/**
 * Makes the ship no longer in supershoot mode.
 */
void Ship::make_unshoot()
{
    _state = _state & ~(PLAYER_SHOOT_BONUS);
    _max_bullets = 1;
}

/**
 * Makes the ship no longer in hyperspace mode.
 */
void Ship::make_unhyper()
{
    _state = _state & ~(PLAYER_HYPER_BONUS);
}

/**
 * Gives the ship the hyperspace bonus.
 */
void Ship::make_hyper()
{
    _hyper_counter = PLAYER_HYPER_TIME;
    _state |= PLAYER_HYPER_BONUS;
}

/**
 * Gives the ship the supershoot bonus.
 */
void Ship::make_shoot()
{
    _shoot_counter = PLAYER_SHOOT_TIME;
    _max_bullets = NUM_BULLETS;
    _state |= PLAYER_SHOOT_BONUS;
}

/**
 * Draws the ship at (x, y). Sets the new image based on the rotation angle
 * of the ship. If the ship is immune, and the immunity counter is an even
 * number, draw the background instead of the ship image. Otherwise, draw the
 * ship image. The will produce a flickering effect for an immune ship, since
 * only every other frame is drawn.
 * @param new_x the new x-coordinate.
 * @param new_y the new y-coordinate.
 */
void Ship::draw(int new_x, int new_y)
{
    SDL_Rect new_rect;

    new_rect.w = _sprite_base->get_width();
    new_rect.h = _sprite_base->get_height();
    new_rect.x = new_x;
    new_rect.y = new_y;

    _picture = _sprite_base->get_images()[_angle / ANGLE_CHANGE];

    if ((_state & PLAYER_IMMUNE) && (_immune_counter % 2 == 0)) {
        update_background();
    } else {
        SDL_BlitSurface(_picture, NULL, _screen, &new_rect);
    }
}

/**
 * Increase the ship's health, never exceeding the maximum allowable health.
 * @param health the amount of health to increase by.
 */
void Ship::give_health(int health)
{
    _health += health;
    if (_health > PLAYER_HEALTH) {
        _health = PLAYER_HEALTH;
    }
}

/**
 * Increment the number of player bullets.
 */
void Ship::incr_bullets()
{
    _bullets++;
}

/**
 * Decrement the number of player bullets.
 */
void Ship::decr_bullets()
{
    if (_bullets > 0) {
        _bullets--;
    }
}

/**
 * Thrust the ship.
 */
void Ship::thrust()
{
    _sprite_base = _base2;
}

/**
 * Unthrust the ship.
 */
void Ship::unthrust()
{
    _sprite_base = _base1;
}

/**
 * Gets the health.
 * @return the health.
 */
const int Ship::get_health()
{
    return _health;
}

/**
 * Gets the state.
 * @return the state.
 */
const int Ship::get_state()
{
    return _state;
}

/**
 * Gets the number of active bullets.
 * @return the number of active bullets.
 */
const int Ship::get_bullets()
{
    return _bullets;
}

/**
 * Gets the maximum number of bullets.
 * @return _max_bullets.
 */
const int Ship::get_max_bullets()
{
    return _max_bullets;
}

/**
 * Gets the current score.
 * @return the current score.
 */
const int Ship::get_score()
{
    return _score;
}

/**
 * Sets the health.
 * @param health the new health.
 */
void Ship::set_health(int health)
{
    _health = health;
}

/**
 * Sets the state.
 * @param the new state.
 */
void Ship::set_state(int state)
{
    _state = state;
}

/**
 * Sets the number of active bullets.
 * @param the new number of active bullets.
 */
void Ship::set_bullets(int bullets)
{
    _bullets = bullets;
}

/** 
 * Sets the maximum number of bullets.
 * @param max_bullets the new maximum number of bullets.
 */
void Ship::set_max_bullets(int max_bullets)
{
    _max_bullets = max_bullets;
}

/**
 * Sets the score.
 * @param the new score.
 */
void Ship::set_score(int score)
{
    _score = score;
}
