/*  Alien Pool
    Copyright (C) 2003  Michael J. Korman <mike@taequin.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/**
 * @file Sprite.cc
 * Implementation of the Sprite class.
 */

#include <cstdlib>
#include <cstdio>
#include <cstring>

#include <string>
#include <fstream>
#include <iostream>
#include <list>

#include <assert.h>

#include <SDL/SDL.h>
#include <SDL_image.h>

#include "Sprite.hh"
#include "Level.hh"
#include "Alien.hh"
#include "Bullet.hh"
#include "common.hh"

extern Alien** aliens;
extern Bullet** alien_bullets;
extern std::list<Level> levels_list;
extern std::list<Level>::iterator current_level;

/**
 * Constructor.
 * @param sb the sprite base.
 * @param scr the screen surface.
 * @param bckgrnd the background surface.
 * @param vel the initial velocity.
 * @param angle the initial angle.
 */
Sprite::Sprite(SpriteBase* sb, SDL_Surface* scr, SDL_Surface* bckgrnd, 
               float xvel, float yvel, int angle)
{
    _sprite_base = sb;
    _xvel = xvel;
    _yvel = yvel;
    _angle = angle;
    _picture = NULL;
    _screen = scr;
    _background = bckgrnd;
}

/**
 * Destructor.
 */
Sprite::~Sprite()
{}

/**
 * Draws the sprite at a new location and clears the old.
 * @param new_x the new x coordinate.
 * @param new_y the new y coordinate.
 */
void Sprite::draw(int new_x, int new_y)
{
    SDL_Rect new_rect;

    new_rect.w = _sprite_base->get_width();
    new_rect.h = _sprite_base->get_height();
    new_rect.x = new_x;
    new_rect.y = new_y;

    _picture = _sprite_base->get_images()[_angle / ANGLE_CHANGE];

    SDL_BlitSurface(_picture, NULL, _screen, &new_rect);
}

/**
 * Redraws the background beneath the sprite.
 */
void Sprite::update_background()
{
    SDL_Rect rect;
    rect.w = _sprite_base->get_width();
    rect.h = _sprite_base->get_height();
    rect.x = _x;
    rect.y = _y;

    SDL_BlitSurface(_background, &rect, _screen, &rect);
}

/**
 * Moves the sprite to (@a x, @a y) and updates the display.
 * @param x the new x-coordinate.
 * @param y the new y-coordinate.
 */
void Sprite::move(int x, int y)
{
    draw(x, y);

    _x = x;
    _y = y;
}

/**
 * Determines if two sprites collide with each other.
 * @param other a pointer to the other sprite.
 * @return @b true if this sprite and @a other collide, @b false otherwise.
 */
const bool Sprite::collide(Sprite* other)
{
    double my_left, its_left;
    double my_right, its_right;
    double my_top, its_top;
    double my_bottom, its_bottom;

    my_left = _x + (_sprite_base->get_width() * 0.1);
    its_left = other->get_x() + (other->get_sprite_base()->get_width() * 0.1);

    my_top = _y + (_sprite_base->get_height() * 0.1);
    its_top = other->get_y() + (other->get_sprite_base()->get_height() * 0.1);

    my_right = _x + (_sprite_base->get_width() * 0.9);
    its_right = other->get_x() + (other->get_sprite_base()->get_width() * 0.9);

    my_bottom = _y + (_sprite_base->get_height() * 0.9);
    its_bottom = other->get_y() + (other->get_sprite_base()->get_height() * 0.9);

    if (my_bottom < its_top) return false;
    if (my_top > its_bottom) return false;
    if (my_right < its_left) return false;
    if (my_left > its_right) return false;

    return true;
}

/**
 * Increments the rotation angle of the sprite.
 */
void Sprite::incr_angle()
{
    _angle += ANGLE_CHANGE;
    
    if (_angle >= 360) {
        _angle = 0;
    }
}

/**
 * Decrements the rotation angle of the sprite.
 */
void Sprite::decr_angle()
{
    _angle -= ANGLE_CHANGE;
    
    if (_angle <= 0) {
        _angle = 360 - ANGLE_CHANGE;
    }
}

/**
 * Determines whether the given coordinates are out-of-bounds.
 * @param x the x-coordinate.
 * @param y the y-coordinate.
 * @return @b true if (@a x, @a y) is out-of-bounds, @b false otherwise.
 */
const bool Sprite::out_of_bounds(int x, int y)
{
    if (x < 0) { 
        return true;
    } else if (x >= (SCREEN_WIDTH - _sprite_base->get_width())) { 
        return true;
    }
    
    if (y < 0) { 
        return true;
    }
    else if (y >= (SCREEN_HEIGHT - _sprite_base->get_height())) { 
        return true;
    }

    return false;
}

/**
 * Gets the sprite base.
 * @return the sprite base.
 */
SpriteBase* Sprite::get_sprite_base() const
{
    return _sprite_base;
}

/**
 * Gets the x-coordinate.
 * @return the x-coordinate.
 */
const int Sprite::get_x()
{
    return _x;
}

/**
 * Gets the y_coordinate.
 * @return the y-coordinate.
 */
const int Sprite::get_y()
{
    return _y;

}

/**
 * Gets the x velocity.
 * @return the x velocity.
 */
const float Sprite::get_xvel()
{
    return _xvel;
}

/**
 * Gets the y velocity.
 * @return the y velocity.
 */
const float Sprite::get_yvel()
{
    return _yvel;
}

/**
 * Gets the current picture.
 * @return the current picture.
 */
const SDL_Surface* Sprite::get_picture()
{
    return _picture;
}

/**
 * Gets the screen surface.
 * @return the screen surface.
 */
const SDL_Surface* Sprite::get_screen()
{
    return _screen;
}

/**
 * Gets the background surface.
 * @return the background surface.
 */
const SDL_Surface* Sprite::get_background()
{
    return _background;
}

/**
 * Gets the angle.
 * @return the angle.
 */
const int Sprite::get_angle()
{
    return _angle;
}

/**
 * Sets the sprite base.
 * @param sb the new sprite base.
 */
void Sprite::set_sprite_base(SpriteBase* sb)
{
    _sprite_base = sb;
}

/**
 * Sets the x-coordinate.
 * @param x the new x-coordinate.
 */
void Sprite::set_x(int x)
{
    _x = x;
}

/**
 * Sets the y-coordinate.
 * @param y the new y-coordinate.
 */
void Sprite::set_y(int y)
{
    _y = y;
}

/**
 * Sets the x velocity.
 * @param vel the x velocity.
 */
void Sprite::set_xvel(float xvel)
{
    _xvel = xvel;
}

/**
 * Sets the y velocity.
 * @param yvel the y velocity.
 */
void Sprite::set_yvel(float yvel)
{
    _yvel = yvel;
}

/**
 * Sets the current picture.
 * @param picture the new picture.
 */
void Sprite::set_picture(SDL_Surface* picture)
{
    _picture = picture;
}

/**
 * Sets the screen surface.
 * @param scrn the new screen surface.
 */
void Sprite::set_screen(SDL_Surface* scrn)
{
    _screen = scrn;
}

/**
 * Sets the background surface.
 * @param bckgrnd the new background surface.
 */
void Sprite::set_background(SDL_Surface* bckgrnd)
{
    _background = bckgrnd;
}

/**
 * Sets the angle.
 * @param angle the new angle.
 */
void Sprite::set_angle(int angle)
{
    _angle = angle;
}
