/*  Alien Pool
    Copyright (C) 2003  Michael J. Korman <mike@taequin.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/**
 * @file SpriteBase.cc
 * Implementation of the SpriteBase class.
 */

#include <string>
#include <iostream>
#include <fstream>

#include <SDL_image.h>

#include "SpriteBase.hh"
#include "common.hh"

/**
 * Constructor. Reads the sprite base information from the @c info file in the
 * sprite's directory. Loads all images into memory.
 * @param dir_name the relative directory name for the sprite.
 * @param screen the screen surface.
 */
SpriteBase::SpriteBase(std::string dir_name, SDL_Surface* screen)
{
    std::string filename("");
    std::string buffer("");
    std::string path("");

    int num_files;

    filename = filename + DATADIR + "/sprites/" + dir_name + "/info";
    std::ifstream infile(filename.c_str());

    if (!infile) {
	std::cerr << "Error opening file " <<  filename << std::endl;
	exit(EXIT_FAILURE);
    }

    std::getline(infile, buffer);
    num_files = atoi(buffer.c_str());

    std::getline(infile, buffer);
    _width = atoi(buffer.c_str());

    std::getline(infile, buffer);
    _height = atoi(buffer.c_str());
    
    for (int i = 0; i < num_files; i++) {
        std::getline(infile, buffer);
        path = path + DATADIR + "/sprites/" + dir_name + "/" + buffer;
	SDL_Surface* image = IMG_Load(path.c_str());

        if (!image) {
            std::cerr << "Error loading image: " << IMG_GetError() << std::endl;
            exit(EXIT_FAILURE);
        }
        
        if (SDL_SetColorKey(image, SDL_SRCCOLORKEY, 
                            SDL_MapRGB(screen->format, 0, 0, 0)) < 0) {
            std::cerr << "Error setting color key: " << SDL_GetError() << std::endl;
            exit(EXIT_FAILURE);
        }
            
        _images.push_back(SDL_DisplayFormat(image));
        SDL_FreeSurface(image);
        path = "";
    }

    _image_count = num_files;
}

/**
 * Destructor. Frees all surfaces associated with this sprite base.
 */
SpriteBase::~SpriteBase()
{
    std::vector<SDL_Surface*>::iterator iter;
    for (iter = _images.begin(); iter != _images.end(); iter++) {
        SDL_FreeSurface(*iter);
    }
}

/**
 * Gets the images vector.
 * @return the images vector.
 */
const std::vector<SDL_Surface*> SpriteBase::get_images()
{
    return _images;
}

/**
 * Gets the image count.
 * @return the image count.
 */
const int SpriteBase::get_image_count()
{
    return _image_count;
}

/**
 * Gets the width of the sprite base.
 * @return the width of the sprite base.
 */
const int SpriteBase::get_width()
{
    return _width;
}

/**
 * Gets the height of the sprite base.
 */
const int SpriteBase::get_height()
{
    return _height;
}

/**
 * Sets the images vector.
 * @param images the new images vector.
 */
void SpriteBase::set_images(std::vector<SDL_Surface*> images)
{
    _images = images;
}

/**
 * Sets the image counter.
 * @param image_counter the new image counter.
 */
void SpriteBase::set_image_count(int image_count)
{
    _image_count = image_count;
}

/**
 * Sets the width of the sprite base.
 * @param width the width of the sprite base.
 */
void SpriteBase::set_width(int width)
{
    _width = width;
}

/**
 * Sets the height of the sprite base.
 * @param height the height of the sprite base.
 */
void SpriteBase::set_height(int height)
{
    _height = height;
}

