/*  Alien Pool
    Copyright (C) 2003  Michael J. Korman <mike@taequin.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/**
 * @file main.cc
 * Contains the main game code.
 */

#include <cstdlib>
#include <cstring>
#include <cmath>
#include <ctime>
#include <iostream>
#include <fstream>
#include <sstream>
#include <list>

#include <assert.h>

#ifdef WIN32
#include <windows.h>
#else
#include <unistd.h>
#endif

#include <SDL/SDL.h>
#include <SDL_image.h>
#include <SDL_ttf.h>
#include <SDL_mixer.h>

#include "Sprite.hh"
#include "SpriteBase.hh"
#include "Ship.hh"
#include "Level.hh"
#include "Bullet.hh"
#include "Alien.hh"
#include "Bonus.hh"
#include "common.hh"

SDL_Surface* background; /**< Background surface. */
SDL_Surface* screen; /**< Screen surface. */
SDL_Surface* main_menu; /**< Main menu surface. */
SDL_Surface* scores_surface; /**< High score surface. */
SDL_Surface* hyperspace_surface; /**< Hyperspace icon surface. */
SDL_Surface* supershoot_surface; /**< Super-shoot icon surface. */

SpriteBase* ship_base; /**< Sprite base for the ship. */
SpriteBase* thrust_ship_base; /**< Sprite base for the thrusting ship. */
SpriteBase* bullet_base; /**< Sprite base for player bullets. */
SpriteBase* alien_base; /**< Sprite base for aliens. */
SpriteBase* dumb_alien_base; /**< Sprite base for dumb aliens. */
SpriteBase* alien_bullet_base; /**< Sprite base for alien bullets. */
SpriteBase* health_bonus_base; /**< Sprite base for health bonuses. */
SpriteBase* shoot_bonus_base; /**< Sprite base for shoot bonuses. */
SpriteBase* hyper_bonus_base; /**< Sprite base for hyperspace bonuses. */

TTF_Font* health_font; /**< Font for "Health:" text. */
TTF_Font* level_font; /**< Font for "Level" text. */
TTF_Font* scores_font; /**< Font for the high scores list. */

SDL_Joystick* joystick; /**< Pointer to the joystick in use. */

Bullet** bullets; /**< Array of player bullets. */
Bullet** alien_bullets; /**< Array of alien bullets. */

Alien** aliens; /**< Array of aliens. */
Ship* ship; /**< Pointer to the ship. */

Bonus* health_bonus; /**< Pointer to the health bonus. */
Bonus* shoot_bonus; /**< Pointer to the shoot bonus. */
Bonus* hyper_bonus; /**< Pointer to the hyperspace bonus. */

Mix_Chunk* health_voice;
Mix_Chunk* shoot_voice;
Mix_Chunk* hyper_voice;
Mix_Chunk* excellent_voice;
Mix_Chunk* gameover_voice;
Mix_Chunk* hyper_sound;
Mix_Chunk* shot_sound;
Mix_Chunk* tock_sound;

Mix_Music* game_music;
Mix_Music* menu_music;
int next_time; /**< Milliseconds that will have passed at the end of this frame. */

GAME_STATE game_state; /**< Current state of the game. */

int level_num; /**< Current level number. */

std::list<Level> levels_list; /**< List of all levels and their attributes. */
std::list<Level>::iterator current_level; /**< Iterator to current level. */

ScoreItem* scores[10];

int new_x; /**< The next x-coordinate of the ship. */
int new_y; /**< The next y-coordinate of the ship. */

// These variables are set when the player inputs a direction.
// A keydown sets them to true, and a keyup sets them back to 
// false. At each frame, the position of the ship will be 
// modified according to the states of these variables. This 
// alleviates the need to worry about an input repeat rate,
// as a move signal will be continously signalled as long as 
// one of these variables is true.
bool move_right;
bool move_left;
bool move_up;
bool move_down;
bool pressed_hyperspace;
bool pressed_thrust;

bool audio_on;

int video_flags;

/**
 * Rounds a number to the nearest integer.
 * @param x number to be rounded.
 * @return the closest integer to @a x.
 */
double round_to_int(double x)
{
    if (x > 0) {
        return int(x + 0.5);
    } else if (x < 0) {
        return int(x - 0.5);
    } else {
        return 0;
    }
}

/**
 * Flushes the event queue.
 */
void flush_event_queue()
{
    SDL_Event event;
    while (SDL_PollEvent(&event));
}

/**
 * Called when the game is over.
 */
void game_over()
{
    SDL_Color color;
    SDL_Surface* text_surface;
    SDL_Rect text_rect;

    text_rect.w = 300;
    text_rect.h = 300;
    text_rect.x = (SCREEN_WIDTH / 2) - 115;
    text_rect.y = (SCREEN_HEIGHT / 2) - 40;

    color.r = 255;
    color.g = 255;
    color.b = 255;

    text_surface = TTF_RenderText_Blended(level_font, "GAME OVER", color);

    SDL_BlitSurface(background, NULL, screen, NULL);
    SDL_BlitSurface(text_surface, NULL, screen, &text_rect);
    SDL_Flip(screen);
    
    if (audio_on) {
        Mix_PlayChannel(5, gameover_voice, 0);
    }

    SDL_Delay(1000);

    if (audio_on) {
        Mix_FadeOutChannel(-1, 1000);
        Mix_FadeOutMusic(1000);
        
        Mix_PlayMusic(menu_music, -1);
    }

    SDL_FreeSurface(text_surface);

    text_rect.w = 700;
    text_rect.h = 300;
    text_rect.x = 195;
    text_rect.y = 300;

    for (int i = 0; i < 10; i++) {
        if (ship->get_score() > scores[i]->score) {
            text_surface = TTF_RenderText_Blended(health_font, 
                                                  "You have achieved a high score!",
                                                  color);
            SDL_BlitSurface(text_surface, NULL, screen, &text_rect);
            SDL_Flip(screen);
            SDL_Delay(2000);

            int limit = i;
            for (int j = 9; j > limit; j--) {
                scores[j]->score = scores[j - 1]->score;
            }
            
#ifdef WIN32
            DWORD size = 255;
            GetUserName(scores[i]->username, &size);
#else
            getlogin_r(scores[i]->username, 255);
#endif
            
            scores[i]->score = ship->get_score();
            break;
        }
    }

    SDL_Delay(1000);
    
    ship->set_score(0);

    game_state = GAME_MENU;

    flush_event_queue();
}

/**
 * Loads the high scores from disk.
 */
void load_scores()
{
    std::string filename(LOCALSTATEDIR "/alienpool.scores");
    std::ifstream infile(filename.c_str());
    std::string tmp_score;
    
    if (!infile) {
        std::cerr << "Error loading high score file " << filename << std::endl;
	for (int i = 0; i < 10; i++) {
	    ScoreItem* score_item = new ScoreItem;
	    char* username = "empty";
            strcpy(score_item->username, username);
	    score_item->score = 0;
	    scores[i] = score_item;
	}
    }

    char username[255];
    int score;
    ScoreItem* score_item;

    int i = 0;
    while ((infile >> username) && (infile >> score)) {
        score_item = new ScoreItem;
        strcpy(score_item->username, username);
        score_item->score = score;
        scores[i] = score_item;
        i++;
    }

    infile.close();
}

/**
 * Writes the high scores to disk.
 */
void write_scores()
{
    std::string filename(LOCALSTATEDIR "/alienpool.scores");
    std::ofstream outfile(filename.c_str());

    if (!outfile) {
        std::cerr << "Error loading high score file " << filename << std::endl;
        return;
    }

    for (int i = 0; i < 10; i++) {
        outfile << scores[i]->username << " " << scores[i]->score << std::endl;
    }
}

/**
 * Displays the high scores.
 */
void display_scores()
{
    SDL_BlitSurface(background, NULL, screen, NULL);

    SDL_Color color;
    color.r = 0;
    color.g = 255;
    color.b = 255;

    SDL_Surface* title_surface = TTF_RenderText_Blended(scores_font,
                                                        "Alien Pool High Scores",
                                                        color);
    SDL_Rect title_rect;
    title_rect.h = 40;
    title_rect.w = 500;
    title_rect.x = 130;
    title_rect.y = 0;
    SDL_BlitSurface(title_surface, NULL, screen, &title_rect);

    for (int i = 0; i < 10; i++) {
        std::string score_str("");
        std::ostringstream score_strm;
        score_strm << scores[i]->score;
        score_str = score_str + scores[i]->username + "......" + score_strm.str();
        
        color.r = 255;
        color.g = 100;
        color.b = 100;
        
        scores_surface = TTF_RenderText_Blended(scores_font, 
                                                score_str.c_str(), color);

        SDL_Rect rect;
        rect.h = 40;
        rect.w = 500;
        rect.x = 0;
        rect.y = (i + 1) * 40 + 10;

        SDL_BlitSurface(scores_surface, NULL, screen, &rect);
    }

    SDL_Flip(screen);
}

/**
 * Performs necessary actions at the start of a level. Displays the 
 * opening text, and initializes sprites.
 */
void start_level()
{
    // Display the introductory level screen.
    SDL_Color color;
    SDL_Surface* text_surface;
    SDL_Rect text_rect;

    text_rect.w = 300;
    text_rect.h = 300;
    text_rect.x = (SCREEN_WIDTH / 2) - 70;
    text_rect.y = (SCREEN_HEIGHT / 2) - 30;

    color.r = 255;
    color.g = 255;
    color.b = 255;

    char level_text[60];
    sprintf(level_text, "Level %d", level_num);
    text_surface = TTF_RenderText_Blended(level_font, level_text, color);

    SDL_BlitSurface(background, NULL, screen, NULL);
    SDL_BlitSurface(text_surface, NULL, screen, &text_rect);
    SDL_Flip(screen);

    SDL_Delay(1000); // Pause while text is displayed.

    SDL_FreeSurface(text_surface);

    // Set the location of the ship, and set its states.
    ship->set_x(SCREEN_WIDTH / 2);
    ship->set_y(SCREEN_HEIGHT / 2);
    ship->set_xvel(0);
    ship->set_yvel(0);
    ship->set_angle(90);
    ship->make_unimmune();

    // Get rid of any bonuses that may be left over from the previous level.
    if (health_bonus) {
        delete health_bonus;
        health_bonus = NULL;
    }
    if (shoot_bonus) {
        delete shoot_bonus;
        shoot_bonus = NULL;
    }
    if (hyper_bonus) {
        delete hyper_bonus;
        hyper_bonus = NULL;
    }

    move_right = false;
    move_left = false;
    move_up = false;
    move_down = false;
    ship->unthrust();

    new_x = SCREEN_WIDTH / 2;
    new_y = SCREEN_HEIGHT / 2;
    
    // Redraw the background.
    SDL_BlitSurface(background, NULL, screen, NULL);

    // Make memory for the aliens.
    aliens = new Alien*[current_level->num_aliens];

    // Create each alien, assigning it a random location on the screen,
    // and adding it to the aliens array.
    Alien* alien;
    for (int i = 0; i < (current_level->num_aliens); i++) {
        alien = new Alien(alien_base, dumb_alien_base, screen, background, 
                          (current_level->alien_change), current_level->alien_change, 0);
        assert(alien);

        // Ensure that no aliens are within 200 units of the ship.
        do {
            alien->set_x((rand() % 556) + 42);
            alien->set_y((rand() % 396) + 42);
        } while (sqrt(pow(alien->get_x() - ship->get_x(), 2) + 
                      pow(alien->get_x() - ship->get_x(), 2)) < 200);

        alien->set_bullet_base(alien_bullet_base);
        aliens[i] = alien;
    }

    // Clear all player bullets.
    for (int i = 0; i < NUM_BULLETS; i++) {
        bullets[i] = NULL;
    }
    ship->set_bullets(0);

    // Clear all alien bullets.
    for (int i = 0; i < NUM_ALIEN_BULLETS; i++) {
        alien_bullets[i] = NULL;
    }

    flush_event_queue();
}

/**
 * Populates the levels list with the data for all levels.
 */
void make_levels()
{
    for (int i = 0; i < 100; i++) {
        Level level;
        level.num_aliens = rand() % 10 + 3;
        level.alien_change = i / 10 + 1;
        level.dumbalien_change = 5 + level.alien_change;
        level.alien_dumb_time = 50;
        level.alien_shoot_time = 250 - (i * 5);
        levels_list.push_back(level);
    }

    current_level = levels_list.begin();
}

/**
 * Determines if any aliens remain.
 * @return @b true if aliens are gone, @b false otherwise.
 */
bool no_more_aliens()
{
    bool found = true;
    for (int i = 0; i < current_level->num_aliens; i++) {
        if (aliens[i] != NULL) {
            found = false;
        }
    }

    return found;
}

/**
 * Initializes the background surface.
 */
void init_bg()
{
    background = IMG_Load(DATADIR "/starscape.png");
    if (!background) {
        std::cerr << "Error loading background: " << IMG_GetError() << std::endl;
        exit(EXIT_FAILURE);
    }

    main_menu = IMG_Load(DATADIR "/mainmenu.png");
    if (!main_menu) {
        std::cerr << "Error loading main menu: " << IMG_GetError() << std::endl;
        exit(EXIT_FAILURE);
    }
}

/**
 * Adds a bullet to the player bullets aray.
 * @param bullet_base the sprite base of the bullet.
 * @param x the x-coordinate of the bullet.
 * @param y the y-coordinate of the bullet.
 * @param trajectory the starting trajectory of the bullet.
 */
void add_bullet(SpriteBase* bullet_base, int x, int y, int trajectory)
{
    int i;

    if (ship->get_bullets() <= (ship->get_max_bullets() - 1)) {
        Bullet* bullet = new Bullet(bullet_base, screen, background, 
                                    BULLET_CHANGE, BULLET_CHANGE, 0);
        for (i = 0; i < NUM_BULLETS; i++) {
            if (bullets[i] == NULL) {
                if (audio_on) {
                    Mix_PlayChannel(7, shot_sound, 0);
                }

                bullet->set_x(x);
                bullet->set_y(y);
                bullet->set_trajectory(trajectory);
                bullets[i] = bullet;
                ship->incr_bullets();

                /* Recoil. */
                ship->set_xvel(ship->get_xvel() - 2*cos((M_PI/180) * float(ship->get_angle())));
                ship->set_yvel(ship->get_yvel() + 2*sin((M_PI/180) * float(ship->get_angle())));
                
                break;
            }
        }
    }
}

/**
 * Loops through all entries in the player bullets array. If the entry is full,
 * move the bullet, and check for collisions.
 */
void manage_bullets()
{
    int i;
    Bullet* bullet;

    for (i = 0; i < NUM_BULLETS; i++) {
        bullet = bullets[i];
	if (bullet != NULL) {
            bullet->go();
            if (bullet->get_ttl() < 0) {
                delete bullet;
                bullets[i] = NULL;
                ship->decr_bullets();
                continue;
            }
            for (int j = 0; j < current_level->num_aliens; j++) {
                if (aliens[j] != NULL) {
                    if (bullet->collide(aliens[j])) {
                        aliens[j]->set_trajectory(bullet->get_trajectory());
                        delete bullet;
                        bullets[i] = NULL;
                        ship->decr_bullets();
                        aliens[j]->hit();
                        break;
                    }
                }
            }
        }
    }
}

/**
 * Loops through all alien bullets. Handles ship collisions and deletion
 * of old bullets.
 */
void manage_alien_bullets()
{
    int i;
    Bullet* bullet;

    for (i = 0; i < NUM_ALIEN_BULLETS; i++) {
        bullet = alien_bullets[i];
	if (bullet != NULL) {
            bullet->go();
            if (bullet->collide(ship)) {
                ship->hit(PLAYER_BULLET_DAMAGE);
                delete bullet;
                alien_bullets[i] = NULL;
                continue;
            }
            if (bullet->get_ttl() < 0) {
                delete bullet;
                alien_bullets[i] = NULL;
                continue;
            }
        }
    }
}

/**
 * Computes the amount of time left for the current frame.
 * @return the number of milliseconds remaining.
 */
int time_left()
{
    int now = SDL_GetTicks();
    if (next_time <= now) {
        return 0;
    } else {
        return next_time - now;
    }
}

/**
 * Updates the on-screen text.
 */
void draw_text()
{
    SDL_Surface* health_surface;
    SDL_Surface* score_surface;
    char health_txt[30];
    char score_txt[30];
    SDL_Color text_color;
 
    text_color.r = 255;
    text_color.g = 255;
    text_color.b = 255;
    sprintf(health_txt, "Health: %d\%%", ship->get_health());
    sprintf(score_txt, "Score: %d", ship->get_score());
    health_surface = TTF_RenderText_Solid(health_font, health_txt, text_color);
    score_surface = TTF_RenderText_Solid(health_font, score_txt, text_color);

    SDL_Rect score_rect;
    score_rect.h = 50;
    score_rect.w = 150;
    score_rect.x = SCREEN_WIDTH - 150;
    score_rect.y = 0;

    SDL_BlitSurface(health_surface, NULL, screen, NULL);
    SDL_BlitSurface(score_surface, NULL, screen, &score_rect);
    SDL_FreeSurface(health_surface);
    SDL_FreeSurface(score_surface);
}

/**
 * Updates the on-screen icons.
 */
void draw_icons()
{
    SDL_Rect hyper_rect;
    hyper_rect.h = 30;
    hyper_rect.w = 30;
    hyper_rect.x = 580;
    hyper_rect.y = 450;

    SDL_Rect super_rect;
    super_rect.h = 30;
    super_rect.w = 30;
    super_rect.x = 610;
    super_rect.y = 450;

    if (ship->get_state() & PLAYER_SHOOT_BONUS)
        SDL_BlitSurface(supershoot_surface, NULL, screen, &super_rect);

    if (ship->get_state() & PLAYER_HYPER_BONUS)
        SDL_BlitSurface(hyperspace_surface, NULL, screen, &hyper_rect);

}

/**
 * If there is no bonus, roll a die to see if we should make one. Otherwise,
 * handle the existing bonus.
 */
void manage_bonus()
{
    // Manage health bonus.
    if (health_bonus == NULL) {
        if (rand() % HEALTH_BONUS_PROB == 1) {
            health_bonus = new Bonus(health_bonus_base, screen, 
                                     background, BONUS_CHANGE, BONUS_CHANGE, 0);
            health_bonus->set_x(rand() % 598);
            health_bonus->set_y(rand() % 438);
        }
    } else {
        if (!health_bonus->go()) {
            delete health_bonus;
            health_bonus = NULL;
        } else {
            health_bonus->move(health_bonus->get_x(), health_bonus->get_y());
            if (health_bonus->collide(ship)) {
                ship->give_health(BONUS_HEALTH);

                if (audio_on) {
                    Mix_PlayChannel(0, health_voice, 0);
                }

                health_bonus->update_background();
                delete health_bonus;
                health_bonus = NULL;
            }
        }
    }

    // Manage shoot bonus.
    if (shoot_bonus == NULL) {
        if (rand() % SHOOT_BONUS_PROB == 1) {
            shoot_bonus = new Bonus(shoot_bonus_base, screen, 
                                     background, BONUS_CHANGE, BONUS_CHANGE, 0);
            shoot_bonus->set_x(rand() % 598);
            shoot_bonus->set_y(rand() % 438);
        }
    } else {
        if (!shoot_bonus->go()) {
            delete shoot_bonus;
            shoot_bonus = NULL;
        } else {
            shoot_bonus->move(shoot_bonus->get_x(), shoot_bonus->get_y());
            if (shoot_bonus->collide(ship)) {
                ship->make_shoot();

                if (audio_on) {
                    Mix_PlayChannel(1, shoot_voice, 0);
                }

                shoot_bonus->update_background();
                delete shoot_bonus;
                shoot_bonus = NULL;
            }
        }
    }

    // Manage hyperspace bonus.
    if (hyper_bonus == NULL) {
        if (rand() % HYPER_BONUS_PROB == 1) {
            hyper_bonus = new Bonus(hyper_bonus_base, screen, 
                                     background, BONUS_CHANGE, BONUS_CHANGE, 0);
            hyper_bonus->set_x(rand() % 598);
            hyper_bonus->set_y(rand() % 438);
        }
    } else {
        if (!hyper_bonus->go()) {
            delete hyper_bonus;
            hyper_bonus = NULL;
        } else {
            hyper_bonus->move(hyper_bonus->get_x(), hyper_bonus->get_y());
            if (hyper_bonus->collide(ship)) {
                ship->make_hyper();

                if (audio_on) {
                    Mix_PlayChannel(2, hyper_voice, 0);
                }

                hyper_bonus->update_background();
                delete hyper_bonus;
                hyper_bonus = NULL;
            }
        }
    }
}

/**
 * Draws the main menu.
 */
void draw_menu()
{
    SDL_BlitSurface(main_menu, NULL, screen, NULL);
    SDL_Flip(screen);
}

/**
 * Initializes game variables and begins executing the main loop.
 */
void main_loop()
{
    Uint8* keys;
    bool done = false;
    SDL_Event event;

    level_num = 1;

    init_bg();

    // Initializes all sprite bases.
    ship_base = new SpriteBase("ship", screen);
    thrust_ship_base = new SpriteBase("thrustship", screen);
    bullet_base = new SpriteBase("bullet", screen);
    alien_base = new SpriteBase("alien", screen);
    dumb_alien_base = new SpriteBase("dumbalien", screen);
    alien_bullet_base = new SpriteBase("alienbullet", screen);
    health_bonus_base = new SpriteBase("healthbonus", screen);
    shoot_bonus_base = new SpriteBase("shootbonus", screen);
    hyper_bonus_base = new SpriteBase("hyperbonus", screen);

    // Allocate space for player and alien bullets.
    bullets = (Bullet**)calloc(NUM_BULLETS, sizeof(Bullet**));
    alien_bullets = (Bullet**)calloc(NUM_ALIEN_BULLETS, sizeof(Bullet**));

    // Create the ship.
    ship = new Ship(ship_base, thrust_ship_base, screen, background, 0, 0, 0);

    // We start off with no bonus.
    health_bonus = NULL;
    shoot_bonus = NULL;
    hyper_bonus = NULL;

    // Populate the levels list.
    make_levels();

    game_state = GAME_MENU;

    if (audio_on) {
        Mix_PlayMusic(menu_music, -1);
        Mix_VolumeMusic(32);
    }

    pressed_hyperspace = false;
    pressed_thrust = false;

    // The main game loop:
    while (!done) {
        if (game_state == GAME_PAUSED) {
            // If the game is paused, don't bother polling for events; just
            // wait for an interrupt.
            SDL_WaitEvent(&event);
            switch (event.type) {
            case SDL_QUIT:
                done = true;
                break;
            case SDL_KEYDOWN:
                switch(event.key.keysym.sym) {
                case SDLK_p: // Un-pause.
                    game_state = GAME_PLAYING;
                    if (audio_on) {
                        Mix_ResumeMusic();
                    }
                    break;
                case SDLK_f: // Make full-screen.
                    video_flags ^= SDL_FULLSCREEN;
                    screen = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, 32, 
           			                          video_flags);
                    break;
                }
            }
        }
        if (game_state == GAME_SCORES) {
            SDL_WaitEvent(&event);
            switch (event.type) {
            case SDL_QUIT:
                done = true;
                break;
            case SDL_KEYDOWN:
                switch (event.key.keysym.sym) {
                case SDLK_f:             
                    video_flags ^= SDL_FULLSCREEN;
                    screen = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, 32, 
           			                          video_flags);
                    break;
                default:
                    game_state = GAME_MENU;
                    break;
                }
                break;
            }
        }
        if (game_state == GAME_MENU) {
            draw_menu();
            SDL_WaitEvent(&event);
            switch (event.type) {
            case SDL_QUIT:
                done = true;
                break;
            case SDL_KEYDOWN:
                switch (event.key.keysym.sym) {
                case SDLK_s:
                    game_state = GAME_PLAYING;
                    current_level = levels_list.begin();
                    level_num = 1;

                    if (audio_on) {
                        Mix_FadeOutMusic(1000);
                        Mix_PlayMusic(game_music, -1);
                    }

                    ship->set_health(PLAYER_HEALTH);
                    ship->make_unhyper();
                    ship->make_unshoot();
                    start_level();
                    SDL_Delay(1000);
                    break;
                case SDLK_f: // Make full-screen.
                    video_flags ^= SDL_FULLSCREEN;
                    screen = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, 32, 
                                              video_flags);
                    break;
                case SDLK_h:
                    game_state = GAME_SCORES;
                    display_scores();
                    break;
                case SDLK_q:
                    if (audio_on) {
                        Mix_FadeOutMusic(1000);
                    }

                    //                    SDL_Delay(1000);
                    done = true;
                    break;
                default:
                    break;
                }
                break;
            }
        }
        if (game_state == GAME_PLAYING) {
            // Set next_time to be the number of milliseconds that will 
            // have elapsed at the end of this frame.
            next_time = SDL_GetTicks() + TICK_INTERVAL;
            while (SDL_PollEvent(&event)) {
                switch (event.type) {
                case SDL_QUIT:
                    done = true;
                    break;
                case SDL_KEYDOWN:
                    switch (event.key.keysym.sym) {
                    case SDLK_q:
                        game_state = GAME_MENU;
                        if (audio_on) {
                            Mix_FadeOutMusic(1000);
                            Mix_PlayMusic(menu_music, -1);
                        }

                        //                        SDL_Delay(1000);
                        break;
                    case SDLK_p:
                        game_state = GAME_PAUSED; // Pause.
                        if (audio_on) {
                            Mix_PauseMusic();
                        }
                        break;
                    case SDLK_SPACE:
                        add_bullet(bullet_base, ship->get_x() + 19, 
                                   ship->get_y() + 19, ship->get_angle());
                        break;
                    case SDLK_h:
                        if (ship->get_state() & PLAYER_HYPER_BONUS) {
                            if (audio_on) {
                                Mix_PlayChannel(6, hyper_sound, 0);
                            }
                            new_x = (rand() % 556) + 42;
                            new_y = (rand() % 396) + 42;
                            ship->make_immune();
                            pressed_hyperspace = true;
                            goto hyperspace;
                        }
                        break;
                    case SDLK_DOWN:
                        move_down = true;
                        break;
                    case SDLK_UP:
                        move_up = true;
                        break;
                    case SDLK_LEFT:
                        move_left = true;
                        break;
                    case SDLK_RIGHT:
                        move_right = true;
                        break;
                    case SDLK_f: // Make full-screen.
                        video_flags ^= SDL_FULLSCREEN;
                        screen = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, 32, 
                                 			      video_flags);
                        break;
                    default:
                        break;
                    }
                    break;
                case SDL_KEYUP:
                    switch (event.key.keysym.sym) {
                    case SDLK_DOWN:
                        move_down = false;
                        break;
                    case SDLK_UP:
                        move_up = false;
                        ship->unthrust();
                        break;
                    case SDLK_LEFT:
                        move_left = false;
                        break;
                    case SDLK_RIGHT:
                        move_right = false;
                        break;
                    default:
                        break;
                    }
                    break;
                case SDL_JOYBUTTONDOWN:
                    if (event.jbutton.button == 0) {
                        add_bullet(bullet_base, ship->get_x() + 19, 
                                   ship->get_y() + 19, ship->get_angle());
                    }
                    if (event.jbutton.button == 3) {
                        if (ship->get_state() & PLAYER_HYPER_BONUS) {
                            new_x = (rand() % 556) + 42;
                            new_y = (rand() % 396) + 42;
                            ship->make_immune();
                            goto hyperspace;
                        }
                    }
                    break;
                case SDL_JOYAXISMOTION:
                    // Joysticks are very sensitive, so we do not want an event
                    // to be registered on every input signal. We therefore set 
                    // a range of (-3200, 3200) to be junk inputs, and everything 
                    // outside that range to be valid axis motion.
                    if (event.jaxis.value < -3200) {
                        if (event.jaxis.axis == 0) {
                            move_left = true;
                        }
                        if (event.jaxis.axis == 1) {
                            move_up = true;
                        }
                    }
                    if (event.jaxis.value > 3200) {
                        if (event.jaxis.axis == 0) {
                            move_right = true;
                        }
                        if (event.jaxis.axis == 1) {
                            move_down = true;
                        }
                    }
                    if (event.jaxis.value < 3200 && event.jaxis.value > -3200) {
                        if (event.jaxis.axis == 0) {
                            move_right = false; move_left = false;
                        }
                        if (event.jaxis.axis == 1) {
                            move_down = false; move_up = false;
                        }

                    }
                    break;
                }
            }
	
            // Check if any of move_down, move_up, move_right, or move_left 
            // are true. If they are, move the ship in the proper direction,
            // or change the rotation angle accordingly.
            if (move_up) {
                ship->set_xvel(ship->get_xvel() + 0.5*cos((M_PI/180) * float(ship->get_angle())));
                ship->set_yvel(ship->get_yvel() - 0.5*sin((M_PI/180) * float(ship->get_angle())));
                ship->thrust();
            }
            if (move_right) {
                ship->decr_angle();
            }
            if (move_left) {
                ship->incr_angle();
            }

        hyperspace:
            // Update the background behind each sprite. We need to clear
            // the backgrounds so that we can redraw them in the proper order. 
            // Note that all backgrounds need to be cleared first, or else 
            // we will see visible boxes around our sprites as updated 
            // backgrounds erase portions of overlapping sprites.
            SDL_Rect rect;
            rect.x = 0;
            rect.y = 0;
            rect.w = SCREEN_WIDTH;
            rect.h = 30;
            // Redraw the background of the text.
            SDL_BlitSurface(background, &rect, screen, &rect);
            rect.x = 0;
            rect.y = 450;
            rect.w = SCREEN_WIDTH;
            rect.h = 30;
            // Redraw the background of the icons.
            SDL_BlitSurface(background, &rect, screen, &rect);

            for (int i = 0; i < NUM_BULLETS; i++) {
                if (bullets[i]) {
                    bullets[i]->update_background();
                }
            }
            for (int i = 0; i < NUM_ALIEN_BULLETS; i++) {
                if (alien_bullets[i]) {
                    alien_bullets[i]->update_background();
                }
            }
            ship->update_background();
            for (int i = 0; i < current_level->num_aliens; i++) {
                if (aliens[i] != NULL) {
                    aliens[i]->update_background();
                }
            }
            if (health_bonus != NULL) {
                health_bonus->update_background();
            }
            if (shoot_bonus != NULL) {
                shoot_bonus->update_background();
            }
            if (hyper_bonus != NULL) {
                hyper_bonus->update_background();
            }

            manage_bullets(); // Manage player bullets.
            manage_alien_bullets(); // Manage alien bullets.
            manage_bonus(); // Manage the bonus.


            if (!pressed_hyperspace) {
                new_x = ship->get_x() + int(ship->get_xvel());
                new_y = ship->get_y() + int(ship->get_yvel());
            }

            ship->go();
            
            if (new_x < 0) {
                new_x = 0;
                ship->set_xvel(ship->get_xvel() * (-1));
            } else if (new_x >= (SCREEN_WIDTH - ship->get_sprite_base()->get_width())) { 
                new_x = SCREEN_WIDTH - ship->get_sprite_base()->get_width();
                ship->set_xvel(ship->get_xvel() * (-1));
            }
            
            if (new_y < 0) { 
                new_y = 0;
                ship->set_yvel(ship->get_yvel() * -1);
            }
            else if (new_y >= (SCREEN_HEIGHT - ship->get_sprite_base()->get_height())) { 
                new_y = SCREEN_HEIGHT - ship->get_sprite_base()->get_height();
                ship->set_yvel(ship->get_yvel() * -1);
            }

            ship->move(new_x, new_y);

            // Loop through the aliens array. If the entry is
            // non-empty, determine if the alien needs to die, and
            // eliminate it if it does. When an alien dies, check if
            // it was the last one, and end the level if it was. If
            // the alien has not died, check if it collided with the
            // ship, and deduct health accordingly. If the alien is
            // dumb, loop through all other aliens and check for
            // collisions. If a dumb alien collides with a normal
            // alien, the normal alien must become dumb.
            for (int i = 0; i < current_level->num_aliens; i++) {
                if (aliens[i] != NULL) {
                    if (!aliens[i]->go()) {
                        ship->set_score(ship->get_score() + 100);
                        delete aliens[i];
                        aliens[i] = NULL;
                        if (no_more_aliens()) {
                            if (audio_on) {
                                Mix_PlayChannel(4, excellent_voice, 0);
                            }
                            delete [] aliens;
                            current_level++;
                            level_num++;
                            start_level();
                            new_x = ship->get_x();
                            new_y = ship->get_y();
                            break;
                        }
                        aliens[i] = NULL;
                        continue;
                    }
                    if (aliens[i]->collide(ship)) {
                        ship->hit(PLAYER_ALIEN_DAMAGE);
                    }
                    if (aliens[i]->get_state() == ALIEN_DUMB) {
                        for (int j = 0; j < current_level->num_aliens; j++) {
                            if (i != j) {
                                if (aliens[j] != NULL && 
                                    aliens[j]->get_state() == ALIEN_NORMAL) {
                                    if (aliens[i]->collide(aliens[j])) {
//                                         if (audio_on) {
//                                             Mix_PlayChannel(8, tock_sound, 0);
//                                         }
                                        int tmp;
                                        tmp = aliens[i]->get_trajectory();
                                        aliens[i]->set_trajectory(aliens[j]->get_trajectory());
                                        aliens[j]->set_trajectory(tmp);
                                        aliens[j]->hit();
                                    }
                                }
                            }
                        }
                    }
                }
            }
            
            draw_text(); // Update the on-screen text.
            draw_icons(); // Update the on-screen icons.

            pressed_hyperspace = false;
            
            SDL_Flip(screen);            
            
            // Idle for the remaining time alloted to this frame. This ensures
            // CPU speed-independent framerates on most systems that can at 
            // least keep up.
            SDL_Delay(time_left());
        }
    }

    // Free objects we have created.
    delete ship;
    delete ship_base;
    delete bullet_base;
    delete alien_base;
    delete dumb_alien_base;
    delete alien_bullet_base;
    delete health_bonus_base;
    delete shoot_bonus_base;
    delete hyper_bonus_base;
    delete health_bonus;
    delete shoot_bonus;
    delete hyper_bonus;
    delete aliens;
}

/**
 * Checks for available joysticks. If any are found, open the first one.
 */
void init_joystick()
{
    std::cout << "Joysticks found: " << SDL_NumJoysticks() << std::endl; 
    if (SDL_NumJoysticks() > 0) {
        std::cout << "#\tName\n";
        std::cout << "-\t----\n";
        for (int i = 0; i < SDL_NumJoysticks(); i++) {
            std::cout << i << "\t" << SDL_JoystickName(i) << std::endl;
        }
        
        SDL_JoystickEventState(SDL_ENABLE);
        joystick = SDL_JoystickOpen(0);
    }
}

/**
 * Initialize fonts.
 */
void init_fonts()
{
    TTF_Init();
    
    health_font = TTF_OpenFont(DATADIR "/VeraBI.ttf", 16);
    level_font = TTF_OpenFont(DATADIR "/VeraBI.ttf", 40);
    scores_font = TTF_OpenFont(DATADIR "/VeraBI.ttf", 30);

    if (!health_font || !level_font) {
        std::cerr << "Error opening font: " << TTF_GetError() << std::endl;
        exit(EXIT_FAILURE);
    }
}

/**
 * Initialize audio.
 */
void init_audio()
{
    if (Mix_OpenAudio(22050, MIX_DEFAULT_FORMAT, 2, 2048) < 0) {
        std::cerr << "Error opening audio: " << SDL_GetError() << std::endl;
        audio_on = false;
    } else {
        audio_on = true;
        
        health_voice = Mix_LoadWAV(DATADIR "/sounds/extrahealth.wav");
        shoot_voice = Mix_LoadWAV(DATADIR "/sounds/supershoot.wav");
        hyper_voice = Mix_LoadWAV(DATADIR "/sounds/hyperspace.wav");
        excellent_voice = Mix_LoadWAV(DATADIR "/sounds/excellentwork.wav");
        gameover_voice = Mix_LoadWAV(DATADIR "/sounds/gameover.wav");
        hyper_sound = Mix_LoadWAV(DATADIR "/sounds/hypersound.wav");
        shot_sound = Mix_LoadWAV(DATADIR "/sounds/shot.wav");
        tock_sound = Mix_LoadWAV(DATADIR "/sounds/tock.wav");
    
        game_music = Mix_LoadMUS(DATADIR "/sounds/gamemusic.ogg");
        menu_music = Mix_LoadMUS(DATADIR "/sounds/menumusic.ogg");
    }
}

/**
 * Load the icons.
 */
void load_icons()
{
    hyperspace_surface = IMG_Load(DATADIR "/h.png");
    supershoot_surface = IMG_Load(DATADIR "/s.png");
}

/**
 * Main function.
 * @param argc number of command line arguments.
 * @param argv array of command line argument strings.
 */
int main(int argc, char* argv[])
{
    if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_JOYSTICK) < 0) {
	std::cerr << "Error loading video: " << SDL_GetError() << std::endl;
	exit(EXIT_FAILURE);
    }
    
    video_flags = SDL_SWSURFACE;// | SDL_DOUBLEBUF;
    screen = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, 32, 
			      video_flags);
			      
    SDL_WM_SetCaption("Alien Pool", NULL);

    SDL_ShowCursor(0); // Hide the mouse pointer.

    if (!screen) {
        std::cerr << "Error creating screen: " << SDL_GetError() << std::endl;
	exit(EXIT_FAILURE);
    }

    srand((unsigned)time(0)); // Seed the random number generator.

    init_joystick();
    init_fonts();
    init_audio();

    load_icons();

    load_scores();

    // Set cleanup callbacks to be executed on exit.
    atexit(SDL_Quit);
    atexit(TTF_Quit);
    atexit(write_scores);

    main_loop(); // Begin the game.

    return 0;
}
